#!/bin/bash

# dyncnn is a deep-learning algorithm for the prediction of
# interacting object dynamics
#
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
# Written by Francois Fleuret <francois.fleuret@idiap.ch>
#
# This file is part of dyncnn.
#
# dyncnn is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# dyncnn is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dyncnn.  If not, see <http://www.gnu.org/licenses/>.

# This script creates the synthetic data-set for shape collision

set -e
set -o pipefail

[[ "${TORCH_NB_THREADS}" ]] || echo "You can set \$TORCH_NB_THREADS to the proper value (default 1)."
[[ "${TORCH_USE_GPU}" ]] || echo "You can set \$TORCH_USE_GPU to 'yes' or 'no' (default 'no')."
[[ "${DYNCNN_DATA_DIR}" ]] || DYNCNN_DATA_DIR="./data/10p-mg"

[[ "${DYNCNN_RUNDIR}" ]] || DYNCNN_RUNDIR="./results"

######################################################################
# Create the data-set if the directory does not exist

if [[ ! -d "${DYNCNN_DATA_DIR}" ]]; then
    cat <<EOF
***************************************************************************
                          Generate the data-set
***************************************************************************
EOF

    make -j -k
    mkdir -p "${DYNCNN_DATA_DIR}"
    # 17 frames every 16 is two frames: t+0, t+16
    ./flatland 40000 \
               --nb_shapes 10 \
               --multi_grasp --every_nth 16 --nb_frames 17 \
               --dir "${DYNCNN_DATA_DIR}"
fi

######################################################################
# Train the model (takes 15h on a GTX 1080 with cuda 8.0, cudnn 5.1,
# and recent torch)

if [[ ! -f "${DYNCNN_RUNDIR}"/model_1000.t7 ]]; then
    cat <<EOF
***************************************************************************
                  Train the model (should take a while)
***************************************************************************
EOF
    ./dyncnn.lua -rundir "${DYNCNN_RUNDIR}"
fi

######################################################################
# Create the images of internal activations using the current.t7 in
# the rundir

cat <<EOF
***************************************************************************
                   Save the internal activation images
***************************************************************************
EOF

for n in 2 12; do
    ./dyncnn.lua -rundir "${DYNCNN_RUNDIR}" -noLog -exampleInternals "${n}"
done

######################################################################
# Plot the loss curves if gnuplot is here

if [[ $(which gnuplot) ]]; then
    cat <<EOF
***************************************************************************
                           Plot the loss curves
***************************************************************************
EOF

    TERMINAL="pdfcairo color transparent enhanced font \"Times,14\""
    EXTENSION="pdf"

    gnuplot <<EOF
set terminal ${TERMINAL}
set output "${DYNCNN_RUNDIR}/losses.${EXTENSION}"
set logscale x
set logscale y
set size ratio 0.75
set xlabel "Number of epochs"
set ylabel "Loss"
plot '< grep "LOSS " "${DYNCNN_RUNDIR}"/log' using 4:6 with l lw 3 lc rgb '#c0c0ff' title 'Validation loss',\
     '< grep "LOSS " "${DYNCNN_RUNDIR}"/log' using 4:5 with l lw 1 lc rgb '#000000' title 'Train loss'

EOF

fi
